#!/usr/bin/env python

import sys
import os
import time
import signal
import subprocess
from decimal import *
import threading
import yaml

def err_msg(s):
	sys.stderr.write(s + '\n')

def prn(s):
	sys.stdout.write(s)
	sys.stdout.flush()

def err_exit(s, ret_v=1):
	err_msg(s)
	sys.exit(ret_v)

hex = lambda s: ''.join( map( lambda c: '{:02x}'.format( ord(c) ), s ) )

map_lst = lambda f, lst: list( map( f, lst ) )
map_up = lambda f, lst: map( lambda a: f(*a), lst )
map_up_lst = lambda f, lst: list( map_up( f, lst ) )

filter_lst = lambda f, lst: list( filter( f, lst ) )
filter_up = lambda f, lst: filter( lambda a: f(*a), lst )
filter_up_lst = lambda f, lst: list( filter_up(f, lst) )

arg_s = lambda k, def_val='': next( ( s[ len(k)+1: ] for s in sys.argv if s.startswith(k+'=') ), def_val )
arg_v = lambda fs, k, def_val: ( lambda s: fs(s) if s else def_val )( arg_s(k) )
arg_i = lambda k, def_val: arg_v( int, k, def_val )
arg_f = lambda k, def_val: arg_v( float, k, def_val )
arg_is = lambda i, s: lst_get(sys.argv, i) == s

def dic_set(d, k, v):
	d[k] = v
	return v

dic_cache = lambda d, k, f_v: d.get(k) if k in d else dic_set( d, k, f_v() )
dic_cache_pop = lambda d, k, def_val=None: d.pop(k) if k in d else def_val

if_not_set = lambda d, k, v: dic_cache(d, k, lambda : v)
pop_if_in = lambda d, k: d.pop(k) if k in d else None

in_range = lambda i, n: 0<=i and i<n
in_wh = lambda x, y, w, h: in_range(x, w) and in_range(y, h)
i_in_lst = lambda i, lst: in_range( i, len(lst) )
lst_get = lambda lst, i, def_val=None: lst[i] if i_in_lst(i, lst) else def_val

###

def cls_to_dic(c):
	ks = filter( lambda k: not k.startswith('_'), dir(c) )
	return dict( map( lambda k: ( k, getattr(c, k) ), ks ) )

def dic_to_cls(d, c=None):
	if not c:
		c = Empty()
	map_lst( lambda kv: setattr(c, *kv), d.items() )
	return c

extends = lambda pcls, c=None: dic_to_cls( cls_to_dic(pcls), c )

class Empty:
	def __init__(self, pcls=None):
		if pcls:
			dic_to_cls( cls_to_dic(pcls), self )
	def __repr__(self):
		if hasattr(self, 'str'):
			return self.str()

get_typ = lambda targ, typ='p': targ.typ if isinstance(targ, Empty) else typ

###

now_sec = time.time
sleep = time.sleep

dec_qt = lambda v, qt='1.00': Decimal(v).quantize( Decimal(qt) )

#str_dt = lambda sec: time.asctime( time.localtime(sec) )
def str_dt(sec):
	t = time.localtime(sec)
	lst = [ t.tm_year, t.tm_mon, t.tm_mday, t.tm_hour, t.tm_min, t.tm_sec ]
	return '{}/{:02}/{:02} {:02}:{:02}:{:02}'.format(*lst)

def str_hms(sec):
	si = int(sec)
	m = si // 60
	s = si % 60 + int( (sec - si) * 100 ) * 0.01
	s = dec_qt(s)
	h = m // 60
	m %= 60
	lst = [ (s, 's') ]
	if m:
		lst.insert( 0, (m, 'm') )
	if h:
		lst.insert( 0, (h, 'h') )
	return ' '.join( map_up( lambda v, s: str(v)+s, lst ) )

###

def new_cnt(n, id_str=''):
	if id_str:
		id_str += ' : '

	e = Empty()
	e.n = n
	e.i = 0
	e.st_sec = None
	e.total_sec = None
	e.show_sec = None
	e.show_interval = 1.0

	def up_no_show(show_fin=True):
		if e.st_sec == None:
			e.st_sec = now_sec()
		else:
			e.i += 1
		run = e.i < e.n
		if not run:
			now = now_sec()
			e.total_sec = now - e.st_sec
			if show_fin:
				show_stat(now)
		return run

	def up(show=True, show_fin=True):
		# show_fin is valid at show == False
		if e.i >= e.n:
			return False
		if not show:
			return up_no_show(show_fin)
		now = now_sec()
		if e.st_sec == None:
			e.st_sec = now
		else:
			e.total_sec = now - e.st_sec
			e.i += 1
		run = e.i < e.n
		if e.total_sec != None and run:
			e.total_sec = e.total_sec * n / e.i
		show_stat(now)
		return run
	e.up = up

	e.cur = lambda : e.i

	def fin_str():
		if e.total_sec == None:
			return 'not start yet'
		if e.i >= n:
			return 'fin ' + str_hms(e.total_sec)
		rest = e.total_sec - ( now_sec() - e.st_sec )
		return 'total {} : rest {} : {}'.format( str_hms(e.total_sec), str_hms(rest), str_dt( e.st_sec + e.total_sec ) )

	pcent = lambda : dec_qt(100*e.i/n, '1.0')
	stat_str = lambda : '{}{}/{}({}%) : {}'.format( id_str, e.i, n, pcent(), fin_str() )

	def show_stat(now=None):
		if not now:
			now = now_sec()
		if e.show_sec == None or e.i >= n or now - e.show_sec >= e.show_interval:
			e.show_sec = now
			print( stat_str() )

	e.get_total_sec = lambda : e.total_sec

	return e

def new_cnt_wait(id_str=''):
	sec = arg_f('sec', 1.0)
	fps = arg_f('fps', 30.0)
	n = arg_i( 'n', int( fps * sec ) )

	cnt = new_cnt(n, id_str)
	e = Empty(cnt)

	e.sec = sec
	e.fps = fps

	e.cur_sec = lambda : sec * cnt.cur() / n

	e.wait_sec = lambda : e.cur_sec() - ( now_sec() - cnt.st_sec )
	def wait():
		w = e.wait_sec()
		if w > 0:
			sleep(w)
	e.wait = wait

	return e

def new_cnt_wh(w, h, id_str=''):
	n = w * h
	cnt = new_cnt(n, id_str)
	e = Empty(cnt)
	e.cur = lambda : ( lambda i: ( i%w, i//w ) )( cnt.cur() )
	return e

###

def cmd_exec(cmd):
	try:
		return subprocess.check_output(cmd, shell=True)
	except:
		return ''

def cmd_proc(cmd):
	try:
		return subprocess.Popen( cmd, shell=True, preexec_fn=os.setsid )
	except:
		return None

def is_alive(proc):
	if not proc:
		return False
	proc.poll()
	return proc.returncode == None

def is_alive_pid(pid):
	if pid < 0:
		return False
	try:
		os.kill(pid, 0)
	except OSError as err:
		return False
	return True

def kill_pid(pid, sig=signal.SIGKILL):
	os.killpg(pid, sig)

def kill_proc(proc, sig=signal.SIGKILL):
	if proc and is_alive(proc):
		kill_pid(proc.pid, sig)

###

exists = lambda path: os.path.exists(path)

def mkdir_if_not(file_path):
	(d, n) = os.path.split(file_path)
	if not exists(d):
		cmd_exec( "mkdir -p '{}'".format(d) )

get_size = lambda path: os.path.getsize(path)

split_ext = lambda path: os.path.splitext(path)

###

def th_start( func, *args ):
	th = threading.Thread( target=func, args=args )
	th.daemon = True
	th.start()
	return th

def new_lock():
	e = Empty()
	e.obj = threading.RLock()
	e.lock = lambda : e.obj.acquire()
	e.unlock = lambda : e.obj.release()
	return e
###

def file_read_inc(fname, def_val='', inc_magic='@include', done=None):
	if done == None:
		done = []
	done.append(fname)

	lst = None
	with open(fname, 'r') as f:
		lst = f.read().split('\n')
	if lst == None:
		return def_val

	def_val = ''
	def func(s):
		if not s.startswith(inc_magic):
			return s
		s = s[ len(inc_magic): ].strip()
		if not s:
			return def_val
		if s[0] in '"<\'':
			s = s[1:-1]
		if s in done:
			return def_val
		return file_read_inc(s, def_val, inc_magic, done)

	return '\n'.join( map( func, lst ) )

def file_write(fname, s):
	try:
		with open(fname, 'w') as f:
			f.write(s)
	except IOError as e:
		err_msg( 'err write {}'.format(fname) )

###

def txt_to_lst_xy(fname):
	lst = []
	with open(fname, 'r') as f:
		lst = f.read().strip().split('\n')
	return map_lst( list, lst )

def lst_xy_wh(lst_xy):
	if not lst_xy:
		return (0, 0)
	ws = map( len, lst_xy )
	w = max(ws) if ws else 0
	return ( w, len(lst_xy) )

lst_xy_get = lambda lst_xy, x, y, def_val=None: lst_get( lst_get( lst_xy, y, [] ), x, def_val )

def lst_xy_set(lst_xy, x, y, v):
	(w, h) = lst_xy_wh(lst_xy)
	if in_wh(x, y, w, h):
		lst_xy[y][x] = v

###

def yaml_load(fname, def_val):
	s = file_read_inc(fname, def_val=None)
	return yaml.load(s) if s != None else def_val

yaml_dump = lambda v, default_flow_style=False: yaml.dump( v, default_flow_style=default_flow_style )

def yaml_save(fname, v, default_flow_style=False):
	file_write( fname, yaml_dump(v) )

def dic_to_items(o):
	magic = '### items ###'
	t = type(o)
	if t == list:
		return map_lst( dic_to_items, o )
	if t == dict:
		ks = sorted( o.keys() )
		return [ magic ] + map_lst( lambda k: [ k, dic_to_items( o.get(k) ) ], ks )
	return o

def items_to_dic(o):
	magic = '### items ###'
	t = type(o)
	if t == list:
		if lst_get(o, 0) == magic:
			return dict( map( lambda kv: ( kv[0], items_to_dic( kv[1] ) ), o[1:] ) )
		return map_lst( items_to_dic, o )
	return o

def try_eval(s, def_val=None):
	try:
		return eval(s)
	except:
		pass
	return def_val

def defs_str_replace(d, top_mark='@', do_eval=False, del_defs=False):
	#	sample yaml
	#
	#	defs:
	#	  foo: abc
	#	  bar: hello
	#	data:
	#	  k1: "@bar world"
	#	  k2: "@foo mart, bar"

	defs_stk = []

	def defs_x(s):
		for d in defs_stk:
			if s not in d:
				continue
			v = d.get(s)
			if type(v) != str:
				return str(v)
			return "'{}'".format(v) if do_eval else v
		return s

	def f(o):
		t = type(o)
		if t == list:
			return map_lst(f, o)
		if t == dict:
			defs = {}
			ks = list( o.keys() )
			if 'defs' in ks:
				defs = o.get('defs')
				if del_defs:
					ks.remove('defs')
				defs_stk.insert(0, defs)
			d = dict( map_lst( lambda k: ( k, f( o.get(k) ) ), ks ) )
			if defs:
				defs_stk.pop(0)
			return d
		if t == str and o.startswith(top_mark):
			s = o[1:] # split top_mark
			bak = ''
			while s != bak:
				bak = s
				s = ' '.join( map( defs_x, s.split(' ') ) )
			return try_eval(s, top_mark+s) if do_eval else s
		return o

	return f(d)

defs_eval = lambda d: defs_str_replace(d, top_mark='*', do_eval=True, del_defs=True)

# EOF
